@@ -1,6 +1,6 @@
 package Language::XSB;
 
-our $VERSION = '0.03';
+our $VERSION = '0.10';
 
 use strict;
 use warnings;
@@ -9,18 +9,25 @@ use Carp;
 
 require Exporter;
 our @ISA = qw(Exporter);
-our %EXPORT_TAGS = ( 'query' => [ qw( xsb_setquery
-				      xsb_clearquery
-				      xsb_query
-				      xsb_next
+our %EXPORT_TAGS = ( 'query' => [ qw( xsb_set_query
+				      xsb_clear_query
 				      xsb_result
+				      xsb_next
+				      xsb_var
+				      xsb_query
 				      xsb_cut
-				      xsb_findall ) ]);
+				      xsb_find_all
+				      xsb_find_one
+				      xsb_call
+				      xsb_assert
+				      xsb_facts ) ]);
 
 our @EXPORT_OK = ( qw(xsb_nreg),
 		   map { @{$EXPORT_TAGS{$_}} } keys(%EXPORT_TAGS));
 our @EXPORT = ();
 
+our (@vars, %vars);
+
 use Language::Prolog::Types qw(F L C isV);
 use Language::XSB::Config;
 use Language::XSB::Base;
@@ -43,6 +50,7 @@ sub callback_perl {
 		    die "subroutine name '$sub' is not a string";
 		UNIVERSAL::isa($args, 'ARRAY') or
 			die "args '$args' is not a list";
+		local (@vars, %vars);
 		# print STDERR "calling sub $sub ( @{$args} )\n";
 		package main;
 		no strict 'refs';
@@ -76,17 +84,22 @@ sub callback_perl {
 sub ok {
     go();
     callback_perl();
-    regtype(1)!=1
+    if ( regtype(1)==1) {
+	@vars=(); %vars=();
+	return 0;
+    }
+    return 1;
 }
 
-sub xsb_setquery (@) {
+sub xsb_set_query (@) {
     defined getreg_int(0) and
 	die "unexpected command sequence";
     while (regtype(1)!=1) {
 	carp "query '".eval{getreg(1)}."' still open, closing";
 	xsb_cut();
     }
-    return grep { isV $_ } @{setreg(1,C(',',@_))};
+    @vars=grep { isV $_ } @{setreg(1,C(',',@_))};
+    return @vars;
 }
 
 sub xsb_query () {
@@ -98,10 +111,25 @@ sub xsb_next () {
 	die "unexpected command sequence";
     regtype(1)==1 and
 	croak "not in a query";
+    %vars=();
     setreg_int(0, 1);
     ok()
 }
 
+sub xsb_var($) {
+    unless (%vars) {
+	defined getreg_int(0) and
+	    die "unexpected command sequence";
+	regtype(1)==1 and
+	    croak "not in a query";
+	@vars{map {$_->name} @vars}=getreg(2)->fargs()
+    }
+    my $name=shift->name;
+    croak "unexistant variable '$name'"
+	unless exists $vars{$name};
+    return $vars{$name}
+}
+
 sub xsb_result () {
     defined getreg_int(0) and
 	die "unexpected command sequence";
@@ -114,7 +142,7 @@ sub xsb_result () {
     getreg(2)->fargs
 }
 
-sub xsb_clearquery () {
+sub xsb_clear_query () {
     defined getreg_int(0) and
 	die "unexpected command sequence";
     regtype(1)==1 and
@@ -132,13 +160,55 @@ sub xsb_cut () {
     ok();
 }
 
-sub xsb_findall (@) {
+
+sub map_vars {
+    return map {
+	isV($_)     ? xsb_var($_) :
+        isL($_)     ? L(_vars(prolog_list2perl_list($_))) :
+        ($_ eq '*') ? xsb_query() :
+	(ref($_) eq '') ? $_ :
+	croak "invalid mapping '$_'";
+    } @_;
+}
+
+
+sub xsb_find_all (@) {
   my @r;
-  xsb_setquery(@_);
-  push (@r, L(xsb_result)) while xsb_next;
+  xsb_set_query(shift);
+  push (@r, map_vars(@_)) while xsb_next;
   return @r
 }
 
+sub xsb_find_one ($;@) {
+    xsb_set_query(shift);
+    if (xsb_next) {
+	my @r=map_vars(@_);
+	xsb_cut;
+	return wantarray ? @r : $r[0];
+    }
+    return ();
+}
+
+sub xsb_call {
+    xsb_set_query(@_);
+    if (xsb_next) {
+	xsb_cut;
+	return 1;
+    }
+    return undef;
+}
+
+sub xsb_assert {
+    my $head=shift;
+    defined $head or croak "xsb_assert called without head";
+    xsb_call F(assertz => C(':-' => $head, C(',', @_)))
+}
+
+sub xsb_facts {
+    return xsb_call C(',', (map { F(assertz => $_) } @_));
+}
+
+
 my $perlcallxsb;
 for my $path (@INC) {
     next if ref $path;
@@ -167,23 +237,24 @@ Language::XSB - use XSB from Perl.
                                 chains=>{plus => '+',
 					 orn => ';'};
 
-    xsb_setquery( equal(X, 34),
-                  equal(Y, -12),
-                  is(Z, plus( X,
-			      Y,
-			      1000 )));
+    xsb_set_query( equal(X, 34),
+                   equal(Y, -12),
+                   is(Z, plus( X,
+		 	       Y,
+			       1000 )));
 
     while(xsb_next()) {
-	printf("X=%d, Y=%d, Z=%d\n", xsb_result())
+	printf("X=%d, Y=%d, Z=%d\n",
+               xsb_var(X), xsb_var(Y), xsb_var(Z))
     }
 
-    print join("\n", xsb_findall(orn(equal(X, 27),
-				     equal(X, 45))));
+    print join("\n", xsb_find_all(orn(equal(X, 27),
+				      equal(X, 45))), X);
 
 =head1 ABSTRACT
 
 Language::XSB provides a bidirectional interface to XSB
-(L<http://xsb.sourceforge.net>).
+(L<http://xsb.sourceforge.net/>).
 
 =head1 DESCRIPTION
 
@@ -215,7 +286,7 @@ package to improve the look of your source (just some syntactic
 sugar).
 
 To make queries to XSB you have to set first the query term with
-the function C<xsb_setquery>, and then use C<xsb_next> and
+the function C<xsb_set_query>, and then use C<xsb_next> and
 C<xsb_result> to iterate over it and get the results back.
 
 Only one query can be open at any time, unless when Perl is
@@ -232,7 +303,7 @@ soubrutines in your script or package:
 
 =over 4
 
-=item C<xsb_setquery(@terms)>
+=item C<xsb_set_query(@terms)>
 
 sets the query term, if multiple terms are passed, then the are
 first chained with the ','/2 functor and the result stored as
@@ -240,35 +311,88 @@ the query.
 
 It returns the free variables found in the query.
 
+
+=item C<xsb_var($var)>
+
+Returns the value binded to C<$var> in the current query/solution combination.
+
+
 =item C<xsb_query()>
 
-returns the current query, variables are bound to its current value if
-C<xsb_next> has been called with success.
+returns the current query, variables are bounded to their current
+values if C<xsb_next> has been called with success.
+
 
 =item C<xsb_next()>
 
 iterates over the query and returns a true value if a new
 solution is found.
 
+
 =item C<xsb_result()>
 
 after calling xsb_next, this soubrutine returns the values
 assigned to the free variables in the query.
 
+
 =item C<xsb_cut()>
 
 ends an unfinished query, similar to XSB (or Prolog) cut
 C<!>. As the real cut in XSB, special care should be taken to
 not cut over tables.
 
-=item C<xsb_findall(@terms)>
 
-does it all in one call (set the query, iterate over it and
-return a list of lists with the results found).
+=item C<xsb_clear_query()>
 
-=item C<xsb_clearquery()>
+a deprecated alias for C<xsb_cut>.
+
+
+=item C<xsb_find_all($query, @pattern)>
+
+iterates over $query and returns and array with @pattern binded to
+every solution. i.e:
+
+  xsb_find_all(member(X, [1, 3, 7, 21]), X)
+
+returns the array C<(1, 3, 7, 21)> and
+
+  xsb_find_all(member(X, [1, 3, 7, 21]), [X])
+
+returns the array C<([1], [3], [7], [21])>.
+
+More elaborate constructions can be used:
+
+  %mothers=xsb_find_all(mother(X,Y), X, Y)
+
+
+=item C<xsb_find_one($query, @pattern>
+
+as C<xsb_find_all> but only for the first solution.
+
+
+=item C<xsb_call>
+
+runs the query once and return true if a solution was found or false
+otherwise.
+
+=item C<xsb_assert($head =E<gt> @body)>
+
+add new definitions at the botton of the database
+
+=item C<xsb_facts(@facts)>
+
+commodity subroutine to add several facts (facts, doesn't have body)
+to the database in one call.
+
+i.e.:
+
+  use Language::Prolog::Sugar functors=>[qw(man woman)];
+
+  xsb_facts( man('teodoro'),
+             man('socrates'),
+             woman('teresa'),
+             woman('mary') );
 
-an alias for C<xsb_cut>.
 
 =back
 
